<?php
/* HTMLConverter - Simple HTML to PDF/RTF/TXT/etc converter
 * Version 1.1.0
 * Copyright 2004-2006, Steve Blinch
 * http://code.blitzaffe.com
 * ============================================================================
 *
 * DESCRIPTION
 *
 * This class takes a pseudo-HTML document as input, and converts it to a
 * formatted plaintext (TXT), Microsoft Word-compatible rich text (RTF), or
 * PDF format document.
 *
 * We refer to the input document as "pseudo-HTML", because the interpreter
 * only handles a VERY limited subset of HTML.  We use it internally to convert
 * articles in our content management system into alternate downloadable 
 * formats, and it serves this purpose well.
 *
 * If your requirements are more complex than <br>, <em>, <strong>, <u>, <p>,
 * and VERY simple CSS, however, then this class probably isn't for you.
 *
 *
 * CHANGELOG
 *
 * 1.1.0 (13-Feb-2007)
 *			- Added support for simple tables in PDF generator (other document
 *			  types still just output one cell per line, however)
 *			- Added support for simple images in PDF generator
 *			- Fixed various bugs
 *
 *
 * USAGE
 *
 * Prior to using this library, you'll need to obtain the ezpdf library from
 * http://www.ros.co.nz/pdf - place this in a subdirectory called "pdf".
 *
 * Then, set the FONT_PATH constant below to point to your PDF font path.
 *
 * Finally, refer to example.php for an example of how to call HTMLConverter.
 *
 *
 * CREDITS
 *
 * For PDF output, this script makes use of the excellent ezpdf library by 
 * Wayne Munro, which is available at: http://www.ros.co.nz/pdf
 *
 * For HTML parsing, this script makes use of the HtmlParser library by Jose
 * Solorzano, which is available at: http://php-html.sourceforge.net/
 *
 *
 * LICENSE
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
 
if (!defined("FONT_PATH")) define("FONT_PATH",dirname(__FILE__)."/fonts/");

require_once(dirname(__FILE__)."/class_DocumentMaker.php");
require_once(dirname(__FILE__)."/htmlparser/htmlparser.inc");
define("HC_DEBUG",false);

class HTMLConverter {

	var $canvas_pixel_width = 1024; // used to convert pixel widths (in tables) to percentages
	var $image_uri_root = '';		// prepended when relative image URLs are encountered

	function reset_state(&$page) {
		$page->state = array(
			"size"=>12,
			"bold"=>false,
			"italic"=>false,
			"underline"=>false,
			"text-align"=>"left",
			"color"=>array(0,0,0)
		);
	}

	// update visual settings based on a "style='...'" property
	function update_style($name,$value,&$page) {
		$this->break_before = false;
		$name = trim($name);
		$value = trim($value);
		switch($name) {
			case "page-break-before":
				if ($value=='always') $this->break_before = true;
				break;
			case "font-size":
				$value = intval($value);
				//echo "update_style setting size to: $value<br>";//fuck
				$page->state["size"] = $value;
				$page->set_font_size($value);
				break;
			case "font-weight":
				if ($value=="bold") $activate = true;
				elseif ($value=="normal") $activate = false;
				$page->bold($activate);
				$page->state["bold"] = $activate;
				break;
			case "font-style":
				if ($value=="italic") $activate = true;
				elseif ($value=="normal") $activate = false;
				$page->italic($activate);
				$page->state["italic"] = $activate;
				break;
			case "text-decoration":
				if ($value=="underline") $activate = true;
				elseif ($value=="none") $activate = false;
				$page->underline($activate);
				$page->state["underline"] = $activate;
				break;
			case "font-color":
				if (substr($value,0,1)=="#") $value = substr($value,1);
				//echo "[$value]";
				$r = hexdec(substr($value,0,2))/255;
				$g = hexdec(substr($value,2,2))/255;
				$b = hexdec(substr($value,4,2))/255;
	
				if (HC_DEBUG) echo "[$r/$g/$b]";
				$page->set_color($r,$g,$b);
				$page->state["color"] = array($r,$g,$b);
				break;
			case "text-align":
				if ($value=="left" || $value=="right" || $value=="center") {
					$page->align($value);
					$page->state["align"] = $value;
					$this->just_explicitly_aligned = true;
				}
				break;
		}
	}

	function parse_style(&$style,&$page) {
		$styles = explode(";",$style);
		foreach ($styles as $k=>$style) {
			$style = trim($style);
			if (!$style) continue;
			list($style_name,$style_value) = explode(":",$style);
			$this->update_style($style_name,$style_value,$page);
		}
	}

	// restore the visual settings for the page to a previously-recorded state
	function restore_state(&$page,$state) {
		if (is_null($state)) return;
		if ($state["italic"]!=$page->state["italic"]) $page->italic($state["italic"]);
		if ($state["bold"]!=$page->state["bold"]) $page->bold($state["bold"]);
		if ($state["size"]!=$page->state["size"]) {
			//echo "<pre>"; var_dump($state); echo "</pre>";
			//echo "restoring size to: ".$state["size"]." (!= ".$page->state["size"].")<br>";//fuck
			$page->set_font_size($state["size"]);
		}
		if ($state["underline"]!=$page->state["underline"]) $page->underline($state["underline"]);
	
		list($er,$eg,$eb) = $page->state["color"];
		list($r,$g,$b) = $state["color"];
		if ($er!=$r || $eg!=$g || $eb!=$b) $page->set_color($r,$g,$b);
		
		if ($state["align"]!=$page->state["align"]) if ($state["align"]) $page->align($state["align"]);
	
		$page->state = $state;
	}

	function push_state(&$page,$tagname,$save_state) {
		$tagname = strtolower($tagname);
		if ($tagname=="br") return;
		if (HC_DEBUG) echo "[Pushing state for $tagname]<br>";
		$state = ($save_state ? $page->state : NULL);
		$page->states[] = array($tagname,$state);
	}

	function preprocess_table($html,$tablewidth,&$page) {
		$row_open = false;
		
		$tables = array();
		$table = array();
		
		if (preg_match_all('/<([\s\S]*?)>/',$html,$matches,PREG_SET_ORDER)) {
			foreach ($matches as $k=>$match) {
				list($tagname,$attributes) = explode(' ',$match[1],2);
				$opening = (substr($tagname,0,1)!='/');
				if (!$opening) $tagname = substr($tagname,1);
//				echo "[$tagname][$attributes]<br>";
				
				switch(strtolower($tagname)) {
					case 'table':
						if ($opening) {
							array_push($tables,$table);
							$table = array();
						} else {
							if (!count($tables)) {
								break 2;
							}
							$table = array_pop($tables);
						}
						break;
					case 'tr':
						$row_open = $opening;
						if (!$opening) {
							if (!count($tables)) {
								break 2;
							}
						}
						$table['columns'] = array();
						break;
					case 'th':
					case 'td':
						if ($opening) {
							if (preg_match('/width\s*?=\s*?\"?\'?([0-9\%]+)/i',$attributes,$matches)) {
								$width = $matches[1];
							} else {
								$width = 0;
							}
							$table['columns'][] = array('width'=>$width);
						}
						break;
				}
			}

			if (substr($tablewidth,-1)=='%') {
				$tablewidth = substr($tablewidth,0,strlen($tablewidth)-1);
			} else {
				$tablewidth = floor($tablewidth * 100 / $this->canvas_pixel_width);
			}
			$table['width'] = $tablewidth;
			$consumedwidth = 0;
			$nowidthcolumns = 0;
			
			foreach ($table['columns'] as $columnno=>$column) {
				$columnwidth = $column['width'];
				if (!$columnwidth) {
					$columnwidth = 0;
					$nowidthcolumns++;
				} elseif (substr($columnwidth,-1)=='%') {
					$columnwidth = substr($columnwidth,0,strlen($columnwidth)-1);
				} else {
					$columnwidth = floor($columnwidth * 100 / $this->canvas_pixel_width);
				}
				$consumedwidth += $columnwidth;
				$table['columns'][$columnno]['width'] = $columnwidth * $tablewidth / 100;
			}
			
			foreach ($table['columns'] as $columnno=>$column) {
				if (!$column['width']) {
					$fullwidth = (100 - $consumedwidth) / $nowidthcolumns;
					$table['columns'][$columnno]['width'] = $fullwidth * $tablewidth / 100;
				}
			}

			return $table;
		}
	}

	function calculate_cell_width(&$table,$cell_number,$colspan) {
		$width = 0;
		for ($i=$cell_number; $i<$cell_number+$colspan; $i++) {
			$width += $table['columns'][$i-1]['width'];
		}
		return $width;
	}
					
	function parse_tag($tag,&$page,$opening,$attrs,&$parser) {
		$tag = strtolower($tag);
		if (!$opening) {
			if ($page->states[count($page->states)-1][0]==$tag) {
				if (HC_DEBUG) echo "[Popping state for $tag]<br>";
				$state = array_pop($page->states);
				$this->restore_state($page,$state[1]);
			} else {
				if (HC_DEBUG) echo "[closing tag: $tag; expecting: ".$page->states[count($page->states)-1][0]."]<br>";
			}
		}
//		$page->text("[$tag]");
		if ($tag!="p") $this->justparagraphed = false;
		
		if ($this->style_tags[$tag]) {
			reset($this->style_tags[$tag]);
			foreach ($this->style_tags[$tag] as $name=>$value) {
				$this->update_style($name,$value,$page);
			}
			// does the above work?  also add support for $this->style_classes
		}
		
		if ($this->break_before) {
			$page->page_break();
			$this->break_before = false;
		}
		
		switch($tag) {
			/*
			case "td":
				if (!$opening) $page->text("\n");
				break;
			*/
			case "table":
				if ($opening) {
					
					$html = substr($parser->iHtmlText,$parser->iHtmlTextIndex);
					$this->current_table = $this->preprocess_table($html,$parser->iNodeAttributes['width'],$page);
					$page->table_begin($this->current_table);
				} else {
					$page->table_end();
				}
				break;
			case "img":
				if ($opening) {
					$src = $parser->iNodeAttributes['src'];
					if (!strpos($src,'://')) $src = $this->image_uri_root . $src;
					$page->image($src,$parser->iNodeAttributes['width'],$parser->iNodeAttributes['align'],$parser->iNodeAttributes['border'],max($parser->iNodeAttributes['hspace'],$parser->iNodeAttributes['vspace']));
				} else {
				}
				break;				
			case "tr":
				if ($opening) {
//					echo "Starting new row<br>";
					$this->row_x = 0;
					$this->cell_number = 1;
					$page->row_begin();
				} else {
//					echo "Ending row<br>";
					$page->row_end();
				}
				break;
			case "th":
				$page->bold($opening);
				$page->state["bold"] = $opening;
			case "td":
				if ($opening) {
					
					if ( (!$this->just_explicitly_aligned) && ($page->state['align']!='left')) $this->update_style('text-align','left',$page);
//					echo "Starting cell at {$this->row_x}<br>";

					$colspan = max((int) $attrs['colspan'],1);
					
					$calculated_width = $this->calculate_cell_width(
						$this->current_table,
						$this->cell_number,
						$colspan
					);
					
					//echo "[$calculated_width]<br>\n"; flush();

					$page->row_cell_begin($this->row_x,$calculated_width);
					
					$this->row_x += $calculated_width;
					
					$this->cell_number += $colspan;
					
				}
				if (!$opening) {
//					echo "Ending cell<br>";
					$page->row_cell_end();
				}
				break;
			case "br":
				$page->text("\n");
				break;
			case "i":
			case "em":
				$page->italic($opening);
				$page->state["italic"] = $opening;
				break;
			case "b":
			case "strong":
				$page->bold($opening);
				$page->state["bold"] = $opening;
				break;
			case "u":
				$page->underline($opening);
				$page->state["underline"] = $opening;
				break;
			case "div":
				$page->text("\n");
				break;
			case "p":
				if ($opening && !$this->justparagraphed) {
					$page->text("\n\n");
					$this->justparagraphed = true;
				}
				break;
		}
	}

	function get_style($style) {
		$res = array();
		$styles = explode(";",trim($style));
		foreach ($styles as $id=>$style) {
			list($name,$value) = explode(":",$style,2);
			$name = trim($name);
			if ($name) $res[$name] = trim($value);
		}
		return $res;
	}
	
	function parse_css($css) {
		if (!$css = trim($css)) return;
		
		// find all individual "tag { css: css; }" blocks
		if (preg_match_all("/([^{]*?){([^}]*?)}[\s]*/i",$css,$matches,PREG_SET_ORDER)) {
			// loop through each block
			foreach ($matches as $id=>$match) {
				// separate the name from the style
				list(,$stylename,$style) = $match;
				$stylename = strtolower(trim($stylename));
				// separate out the style information
				$style = $this->get_style($style);
				
				// figure out what kind of style it is, and add it to
				// the appopriate array
				if (substr($stylename,0,1)==".") {
					$this->style_classes[substr($stylename,1)] = $style;
				} elseif (substr($stylename,0,1)=="#") {
				} else {
					$this->style_tags[$stylename] = $style;
				}
			}
		}
	}

	function convert_file($filename,$cssfile,&$docmaker,$initialize=false,$header=false,$footer=false) {
		$document = implode("",file($filename));
		$css = implode("",file($cssfile));
		$this->convert($document,$css,$docmaker,$initialize,$header,$footer);
	}

	function convert($document,$css,&$docmaker,$initialize=false,$header=false,$footer=false) {
		$page = &$docmaker;

		$document = str_replace("\r\n"," ",$document);
		$document = str_replace("\n"," ",$document);
		$document = preg_replace("/[\s]{2,}/"," ",$document);

		// grab any CSS that may appear in a <style> block in the document itself;
		// add this to the CSS provided by the user if found
		if (preg_match("/<style[^>]*type=['\"]?text\/css['\"]?[^>]*>([\s\S]*)<\/style>/i",$document,$matches)) {
			$css .= $matches[1];
		}
		$this->parse_css($css);
		
		// Get rid of a few things we don't need (eg: everything but the contents of the
		// <body> tag.  We don't just preg_match() the contents of <body>...</body>, though,
		// in case a plain block of HTML has been passed (without a surronding <body> tag).
		$document = preg_replace(
			array(
				"/<!DOCTYPE([\s\S]*?)>/i",
				"/<\?xml([\s\S]*?)\?>/i",
				"/<head>([\s\S]*)<\/head>/i",
				"/<\/?html([\s\S]*?)>/i",
				"/<\/?body([\s\S]*?)>/i"
			),
			"",
			$document
		);
		
		$font_size = 12;
		if ($initialize) {
			$page->prepare_fonts(array("Times New Roman","Helvetica"),1,$font_size);
			$page->document_begin("portrait");
			$page->set_margins(0.75,0.75,0.75,0.75);
			$page->line_spacing(1);
		}
		$page->no_starting_blanks = true;

		if (!$header) {
			$header = array(
				"center"=>array(
					"text"=>"Imported HTML Document",
					"font"=>1,
					"size"=>12,
					"bold"=>true,
					"italic"=>false,
					"underline"=>false
				),
			);
			$page->header($header);
		}
		
		if (!$footer) {
			
			$footer = array(
				"left"=>array(
					"text"=>"Converted by DocMaker",
					"font"=>1,
					"size"=>12,
					"bold"=>false,
					"italic"=>false,
					"underline"=>false
				),
				"center"=>array(
					"text"=>"Page {PAGENUM} of {TOTALPAGENUM}",
					"font"=>1,
					"size"=>12,
					"bold"=>false,
					"italic"=>false,
					"underline"=>false
				),
				"right"=>array(
					"text"=>"http://code.blitzaffe.com",
					"font"=>1,
					"size"=>12,
					"bold"=>false,
					"italic"=>false,
					"underline"=>false
				),
			);
			$page->footer($footer);
		}
		
		$this->reset_state($page);
		$page->state["size"] = $font_size;

		$parser = &new HtmlParser ($document);
		while ($parser->parse()) {
		     
			switch($parser->iNodeType) {
				case NODE_TYPE_START:
					if (HC_DEBUG) echo "Starting..<BR>";
					break;
				case NODE_TYPE_ELEMENT:
					$attrs = &$parser->iNodeAttributes;
		
					$this->push_state($page,$parser->iNodeName,true); //is_array($attrs["style"]) || $attrs["class"] || );

					$this->just_explicitly_aligned = false;
					
					
					if ($attrs["class"]) {
						$styleclass = strtolower($attrs["class"]);
						if ($this->style_classes[$styleclass]) {
							reset($this->style_classes[$styleclass]);
							foreach ($this->style_classes[$styleclass] as $name=>$value) {
								$this->update_style($name,$value,$page);
							}
						}
					}
					if ($attrs["style"]) {
						$this->parse_style($attrs["style"],$page);
					}
					
					$this->parse_tag($parser->iNodeName,$page,true,$attrs,$parser);

					
					//echo "Element: name=".$parser->iNodeName.",value=".$parser->iNodeValue.",attrs=".$parser->iNodeAttributes."<br>";
					break;
				case NODE_TYPE_ENDELEMENT:
					$attrs = &$parser->iNodeAttributes;

					$this->parse_tag($parser->iNodeName,$page,false,$attrs,$parser);
					//echo "EndElement: name=".$parser->iNodeName.",value=".$parser->iNodeValue.",attrs=".$parser->iNodeAttributes."<br>";
					break;
				case NODE_TYPE_TEXT:
					$text = $parser->iNodeValue;
					if (trim($text)) $page->text(html_entity_decode($text));
					$this->justparagraphed = false;
					break;
				case NODE_TYPE_COMMENT:
					break;
				case NODE_TYPE_DONE:
					if (HC_DEBUG) echo "Ending..<BR>";
					break;
		    }
		}  

		if ($initialize) $page->document_end();
	}
}
?>